#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Discrepancy class for Bayesian inference method

Option A:
    With no explicitly-specified model

Option B:
    With known redidual variance sigma2 (Gaussian or others)

Option C:
    With unknown residual variance sigma2 (Gaussian)
    with given distribution

Author: Farid Mohammadi, M.Sc.
E-Mail: farid.mohammadi@iws.uni-stuttgart.de
Department of Hydromechanics and Modelling of Hydrosystems (LH2)
Institute for Modelling Hydraulic and Environmental Systems (IWS), University
of Stuttgart, www.iws.uni-stuttgart.de/lh2/
Pfaffenwaldring 61
70569 Stuttgart

Created on Mon Sep  2 10:48:35 2019
"""
import scipy.stats as stats
from surrogate_models.exp_designs import ExpDesigns
from surrogate_models.inputs import Input


class Discrepancy:
    def __init__(self, InputDisc):
        self.type = 'Gaussian'
        self.parameters = None
        self.name = 'Sigma2'
        self.prior = None
        self.Marginals = []
        self.InputDisc = InputDisc
        self.n_samples = 10000
        self.sigma2_prior = None

    def create_inputDisc(self):
        InputClass = Input()
        InputClass.addMarginals()

        return InputClass

    # -------------------------------------------------------------------------
    def get_sample(self, n_samples):
        """
        Generate samples for the sigma2s, i.e. the diagonal entries of the
        variance-covariance matrix in the multivariate normal distribution.

        Parameters
        ----------
        n_samples : int
            Number of samples (parameter sets).

        Returns
        -------
        sigma2_prior: array of shape (n_samples, n_params)
            Sigma2 samples.

        """
        self.n_samples = n_samples
        ExpDesign = ExpDesigns(self.InputDisc)
        self.sigma2_prior = ExpDesign.generate_ED(
            n_samples, sampling_method='random', max_pce_deg=1
            )
        # Store BoundTuples
        self.ExpDesign = ExpDesign

        # Naive approach: Fit a gaussian kernel to the provided data
        self.ExpDesign.JDist = stats.gaussian_kde(ExpDesign.raw_data)

        # Save the names of sigmas
        if len(self.InputDisc.Marginals) != 0:
            self.name = []
            for Marginalidx in range(len(self.InputDisc.Marginals)):
                self.name.append(self.InputDisc.Marginals[Marginalidx].name)

        return self.sigma2_prior
