#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Fri Aug  9 16:25:43 2019
This example is for the calibration and validation scenario.
@author: farid
    """

import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
plt.style.use('seaborn-white')
plt.rcParams.update({'font.size': 18})
plt.rc('font', family='sans-serif', serif='Arial')
plt.rc('figure', figsize = (24, 16))
plt.rc('text', usetex=True)
plt.rc('xtick', labelsize=18)
plt.rc('ytick', labelsize=18)
plt.rc('axes', labelsize=18)

import seaborn as sns
import os

try:
    import cPickle as pickle
except ModuleNotFoundError:
    import pickle

from PyLink.FuncForwardModel import FuncForwardModel
from surrogate_models.Input import Input
from surrogate_models.surrogate_models import Metamodel
from PostProcessing.PostProcessing import PostProcessing
from BayesInference.BayesInference import BayesInference, Discrepancy

if __name__ == "__main__":
    
    #=====================================================
    #=============   COMPUTATIONAL MODEL  ================
    #=====================================================
    Model = FuncForwardModel()
    
    Model.Type = 'Function'
    Model.pyFile = 'AnalyticalFunction'
    Model.Name = 'AnalyticFunc'
    
    Model.Output.Names = ['Z']
    
    
    # For Bayesian inversion synthetic data with X=[0,0]
    Model.Observations['Time [s]'] = np.arange(0, 5, 1.) / 5
    Model.Observations['Z'] = np.repeat([2],5)
    
    # For Checking with the MonteCarlo refrence
#    Model.MCReference['Time [s]'] = np.arange(0, 10, 1.) / 9
#    Model.MCReference['mean'] = np.array([127.13880108, 127.12652988, 127.12144698,
#                             127.11754674, 127.11425868, 127.11136184, 127.1087429,
#                             127.10633454, 127.10409289, 127.10198748])
#    Model.MCReference['std'] = np.array([171.2520775,  171.21550753, 171.20352691,
#                             171.19559223, 171.18975202, 171.18525035, 171.18169953,
#                             171.17886925, 171.17660944, 171.17481586])

    # For Bayesian inversion synthetic data with X=[1,1]
    Model.ObservationsValid['Time [s]'] = np.arange(0, 5, 1.) / 5
    Model.ObservationsValid['Z']= np.array([2.21773563, 2.11712764, 
                             2.02460905, 1.93849485, 1.85761462])
    
    #=====================================================
    #=========   PROBABILISTIC INPUT MODEL  ==============
    #=====================================================
    # Define the uncertain parameters with their mean and 
    # standard deviation
    Inputs = Input()
    
    ndim = 2
    for i in range(ndim):
        Inputs.addMarginals()
        Inputs.Marginals[i].Name = '$X_{%s}$'%(i+1)
        Inputs.Marginals[i].DistType = 'unif'
        Inputs.Marginals[i].Parameters =  [-5, 5]
    
    #=====================================================
    #==========  DEFINITION OF THE METAMODEL  ============
    #=====================================================    
    MetaModelOpts = Metamodel(Inputs)
    
    # Select if you want to preserve the spatial/temporal depencencies
    MetaModelOpts.DimRedMethod = 'PCA'
    
    # Select your metamodel method
    # 1) PCE (Polynomial Chaos Expansion) 2) GPE (Gaussian Process Emulator)
    # 3) PCEKriging (PCE + GPE)
    MetaModelOpts.metaModel = 'GPE'
    
    # ------------------------------------------------
    # ------------- PCE Specification ----------------
    # ------------------------------------------------
    # Select the sparse least-square minimization method for 
    # the PCE coefficients calculation:
    # 1)OLS: Ordinary Least Square  2)BRR: Bayesian Ridge Regression 
    # 3)LARS: Least angle regression  4)ARD: Bayesian ARD Regression
    # 5)FastARD: Fast Bayesian ARD Regression
    # 6)SGDR: Stochastic gradient descent learning
    MetaModelOpts.RegMethod = 'FastARD'
    
    # Specify the max degree to be compared by the adaptive algorithm:
    # The degree with the lowest Leave-One-Out cross-validation (LOO)
    # error (or the highest score=1-LOO)estimator is chosen as the final 
    # metamodel.
    MetaModelOpts.MinPceDegree = 1 #3
    MetaModelOpts.MaxPceDegree = 8 #15
    
    # q-quasi-norm 0<q<1 (default=1)
    MetaModelOpts.q = 1 #0.75
    
    # Print summary of the regression results
    #MetaModelOpts.DisplayFlag = True
    
    # ------------------------------------------------
    # ------ Experimental Design Configuration -------
    # ------------------------------------------------
    
    # Generate an experimental design of size NrExpDesign based on a latin 
    # hypercube sampling of the input model or user-defined values of X and/or Y:
    MetaModelOpts.addExpDesign()
    
    # One-shot (normal) or Sequential Adaptive (sequential) Design
    MetaModelOpts.ExpDesign.Method = 'normal'
    NrofInitSamples = 200 #75
    MetaModelOpts.ExpDesign.NrSamples = NrofInitSamples
    
    # Sampling methods
    # 1) random 2) latin_hypercube 3) sobol 4) halton 5) hammersley 6) chebyshev(FT) 
    # 7) korobov 8) grid(FT) 9) nested_grid(FT) 10)user
    MetaModelOpts.ExpDesign.SamplingMethod = 'random'
    
    # Sequential experimental design (needed only for sequential ExpDesign)
    MetaModelOpts.ExpDesign.NrofNewSample = 1
    MetaModelOpts.ExpDesign.MaxNSamples = 50 #150
    MetaModelOpts.ExpDesign.ModifiedLOOThreshold = 1e-16
    
    # Defining the measurement error, if it's known a priori
    obsData = pd.DataFrame(Model.Observations, columns=Model.Output.Names)
    DiscrepancyOpts = Discrepancy('')
    DiscrepancyOpts.Type = 'Gaussian'
    DiscrepancyOpts.Parameters = obsData  ** 2
    MetaModelOpts.Discrepancy = DiscrepancyOpts
        
    # Plot the posterior snapshots for SeqDesign
    MetaModelOpts.ExpDesign.PostSnapshot = False
    MetaModelOpts.ExpDesign.stepSnapshot = 1
    MetaModelOpts.ExpDesign.MAP = (0,0)
    MetaModelOpts.ExpDesign.parNames = ['$X_1$', '$X_2$']
    
    # ------------------------------------------------
    # ------- Sequential Design configuration --------
    # ------------------------------------------------
    # 1) 'None' 2) 'epsilon-decreasing'
    MetaModelOpts.ExpDesign.TradeOffScheme = 'epsilon-decreasing'
    #MetaModelOpts.ExpDesign.nReprications = 2
    # -------- Exploration ------
    #1)'Voronoi' 2)'MC' 3)'LHS' 4)'dual annealing'
    MetaModelOpts.ExpDesign.ExploreMethod = 'Voronoi'
    
    # Use when 'dual annealing' chosen
    MetaModelOpts.ExpDesign.MaxFunItr = 200
    
    # Use when 'MC' or 'LHS' chosen
    MetaModelOpts.ExpDesign.NCandidate = 100
    MetaModelOpts.ExpDesign.NrofCandGroups = 4
    
    # -------- Exploitation ------
    # 1)'BayesOptDesign' 2)'ActiveLearning' 3)'alphabetic' 4)'None' (Space-filling)
    MetaModelOpts.ExpDesign.ExploitMethod = 'None'
    
    # BayesOptDesign -> when data is available
    # 1)DKL (Kullback-Leibler Divergence) 2)DPP (D-Posterior-percision)
    # 3)APP (A-Posterior-percision) 
    MetaModelOpts.ExpDesign.UtilityFunction = 'DKL' #['DKL', 'DPP']
    
    # ActiveLearning
    # 1)ALM (Active learning MacKay) 2)ALC (Active learning Cohn)
    # 3) 
    #MetaModelOpts.ExpDesign.UtilityFunction = 'ALM' #['ALM', 'ALC']
    
    # alphabetic
    # 1)D-Opt (D-Optimality) 2)A-Opt (A-Optimality)
    # 3)K-Opt (K-Optimality)
    # MetaModelOpts.ExpDesign.UtilityFunction = 'D-Opt' #['D-Opt', 'A-Opt', 'K-Opt']  
    
    # >>>>>>>>>>>>>>>>>>>>>> Build Surrogate <<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<
    # MetaModelOpts.slicingforValidation = True
    MetaModelOpts.index = 5
    # Adaptive sparse arbitrary polynomial chaos expansion
    # PCEModelCalib, PCEModelValid = MetaModelOpts.create_metamodel(Model)
    PCEModel = MetaModelOpts.create_metamodel(Model)
    
    #=====================================================
    #=========  POST PROCESSING OF METAMODELS  ===========
    #=====================================================
    #PostPCE = PostProcessing(PCEModelCalib)
    PostPCE = PostProcessing(PCEModel)
    
    PostPCE.Name = 'Calib'
    PostPCE.validMetamodel(nValidSamples= 3)
    
    # Compute the moments and compare with the Monte-Carlo reference
#    PostPCE.plotMoments()
    
    if MetaModelOpts.metaModel != 'GPE':
        # Plot the evolution of the KLD,BME, and Modified LOOCV error
        if MetaModelOpts.ExpDesign.Method == 'sequential':
            PostPCE.seqDesignDiagnosticPlots()
        
        # Plot the sobol indices
        sobol_cell, total_sobol = PostPCE.sobolIndicesPCE()
    
    # Compute and print RMSE error
    PostPCE.accuracyCheckMetaModel(nSamples=3000)
    
    
    #=====================================================
    #=======  Bayesian calibration with Emulator =========
    #=====================================================
    # BayesOpts = BayesInference(PCEModelCalib)
    BayesOpts = BayesInference(PCEModel)
    BayesOpts.Name = 'Calib'
    BayesOpts.emulator = True
    
    # BME Bootstrap
    BayesOpts.Bootstrap = True
    BayesOpts.NrofSamples = 100000
    BayesOpts.BootstrapItrNr = 10
    BayesOpts.BootstrapNoise = 0.05
    
    # Select the inference method
    BayesOpts.SamplingMethod = 'MCMC'
    BayesOpts.MCMCnSteps = 1000
    BayesOpts.MCMCnwalkers = 50
    BayesOpts.MultiProcessMCMC = True
    
    BayesOpts.PlotPostDist = True
    BayesOpts.PlotPostPred = True
    
    # ----- Define the discrepancy model -------
    obsData = pd.DataFrame(Model.Observations, columns=Model.Output.Names)
    # (Option B)
    DiscrepancyOpts = Discrepancy('')
    DiscrepancyOpts.Type = 'Gaussian'
    DiscrepancyOpts.Parameters = obsData[:MetaModelOpts.index]**2
    
    BayesOpts.Discrepancy = DiscrepancyOpts
    
    # (Option C)
#    DiscInputOpts = Input()
#    
#    DiscInputOpts.addMarginals()
#    DiscInputOpts.Marginals[0].Name = 'Sigma2'
#    DiscInputOpts.Marginals[0].DistType = 'unif'
#    DiscInputOpts.Marginals[0].Parameters =  [0, 5e-1]
#    
#    DiscrepancyOpts = Discrepancy(DiscInputOpts)
#    
#    BayesOpts.Discrepancy = DiscrepancyOpts
    
    
    Bayes_Calib = BayesOpts.create_Inference()
    
    # Make directory for saving the plots/outputs
    newpath = (r'Outputs_'+MetaModelOpts.ExpDesign.Method) 
    if not os.path.exists(newpath): os.makedirs(newpath)

    
    ## Plot the posterior (Model)
    with sns.axes_style("white"):
        g = sns.jointplot(x=Bayes_Calib.Posterior_df['$X_{1}$'], y=Bayes_Calib.Posterior_df['$X_{2}$'], 
                          kind="kde", cmap='jet');
        g.ax_joint.collections[0].set_alpha(0)
        g.set_axis_labels("$X_1$", "$X_2$")
        
        g.savefig('./'+newpath+'/Posterior_PCEModel_Calib_'+MetaModelOpts.ExpDesign.Method+'.svg')
        plt.close()
        
    
    #=====================================================
    #=======  Bayesian validation with Emulator =========
    #=====================================================
    # BayesOpts_Valid = BayesInference(PCEModelValid)
    BayesOpts_Valid = BayesInference(PCEModel)
    
    BayesOpts_Valid.Name = 'Valid'
    BayesOpts_Valid.emulator = True
    
    # BME Bootstrap
    BayesOpts_Valid.Bootstrap = True
    BayesOpts_Valid.MCMCinitSamples = Bayes_Calib.Posterior_df
    BayesOpts_Valid.BootstrapItrNr = 10
    BayesOpts_Valid.BootstrapNoise = 0.05
    
    # Select the inference method
    BayesOpts_Valid.SamplingMethod = 'MCMC'
    BayesOpts_Valid.MCMCnSteps = 1000
    BayesOpts_Valid.MCMCnwalkers = 50
    BayesOpts.MultiProcessMCMC = True
    
    BayesOpts_Valid.PlotPostDist = True
    BayesOpts_Valid.PlotPostPred = True
    
    # ----- Define the discrepancy model -------
    obsDataValid = pd.DataFrame(Model.ObservationsValid, columns=Model.Output.Names)
    # (Option B)
    DiscrepancyOpts = Discrepancy('')
    DiscrepancyOpts.Type = 'Gaussian'
    DiscrepancyOpts.Parameters = 0.01 * obsDataValid**2
    
    BayesOpts_Valid.Discrepancy = DiscrepancyOpts
    
    Bayes_Valid = BayesOpts_Valid.create_Inference()
    
    
    ## Plot the posterior (PCEModel)
    with sns.axes_style("white"):
        g = sns.jointplot(x=Bayes_Valid.Posterior_df['$X_{1}$'], y=Bayes_Valid.Posterior_df['$X_{2}$'], 
                          kind="kde", cmap='jet');
        g.ax_joint.collections[0].set_alpha(0)
        g.set_axis_labels("$X_1$", "$X_2$")
        g.savefig('./'+newpath+'/Posterior_PCEModel_Valid_'+MetaModelOpts.ExpDesign.Method+'.svg')
        plt.close()
    
    
    #=====================================================
    #==============  Save class objects  =================
    #=====================================================
    with open('AnalyticFunc_Results.pkl', 'wb') as output:
        pickle.dump(PCEModel, output, pickle.HIGHEST_PROTOCOL)
        
        # pickle.dump(PCEModelValid, output, pickle.HIGHEST_PROTOCOL)
        
        pickle.dump(PostPCE, output, pickle.HIGHEST_PROTOCOL)
    
        pickle.dump(Bayes_Calib, output, pickle.HIGHEST_PROTOCOL)
        
        pickle.dump(Bayes_Valid, output, pickle.HIGHEST_PROTOCOL)
    
#    del PCEModel
#    del PostPCE
#    del Bayes
    
    # Load the objects
#    with open('AnalyticFunc_Results.pkl', 'rb') as input:
#        PCEModel = pickle.load(input)
#        PostPCE = pickle.load(input)
#        Bayes_PCE = pickle.load(input)    
#        Bayes_Model = pickle.load(input) 