#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
This test deals with a surrogate-assisted Bayesian calibration of a two-phase 
    CO2 storage model¹.


You will see how to:
    Construct a surrogate model
    Perform sensitivity analysis via Sobol Indices
    Calibration via Bayesian inference
    
Author: Farid Mohammadi, M.Sc.
E-Mail: farid.mohammadi@iws.uni-stuttgart.de
Department of Hydromechanics and Modelling of Hydrosystems (LH2)
Institute for Modelling Hydraulic and Environmental Systems (IWS), University of Stuttgart
www.iws.uni-stuttgart.de/lh2/
Pfaffenwaldring 61
70569 Stuttgart

1: Köppel M, Franzelin F, Kröker I, Oladyshkin S, Santin G, Wittwar D, u. a. 
   Comparison of data-driven uncertainty quantification methods for a carbon 
   dioxide storage benchmark scenario. Computational Geosciences. 2019;23(2):339–54.
"""
import sys
import os
import numpy as np
import pandas as pd
import seaborn as sns
import matplotlib
matplotlib.use('agg')

sys.path.insert(0,'./../../../BayesValidRox')

try:
    import cPickle as pickle
except ModuleNotFoundError:
    import pickle
    

from PyLink.PyLinkForwardModel import PyLinkForwardModel
from surrogate_models.Input import Input
from surrogate_models.surrogate_models import Metamodel
from PostProcessing.PostProcessing import PostProcessing
from BayesInference.BayesInference import BayesInference, Discrepancy



if __name__ == "__main__":
    
    refDataDir = './refData/'
    #=====================================================
    #=============   COMPUTATIONAL MODEL  ================
    #=====================================================
    Model = PyLinkForwardModel()
    
    Model.Type = 'PyLink'
    Model.Name = 'CO2Benchmark'
    Model.InputFile = "interface.xml"
    Model.InputTemplate = "interface.tpl.xml"
    
    Model.nrCPUs = 4
    Model.Command = "model_CO2.py -f interface.xml"
    Model.ExecutionPath = os.getcwd()
    Model.Output.Parser = 'read_CO2'
    Model.Output.Names = ['Pressure [bar]']
    Model.Output.FileNames = ["PressureOutput.csv"]
    
    # For Bayesian inversion 
    # MAP = (1.43e-13, 0.259, 1.09e-12)
    Model.MeasurementFile = 'MeasuredPressure.csv'
    
    # For Checking with the MonteCarlo refrence
    Model.MCReference = True
    Model.MCReferenceFile = 'MCrefs_MeanStd.csv'
    
    #=====================================================
    #=========   PROBABILISTIC INPUT MODEL  ==============
    #=====================================================
    Inputs = Input()
    
    # Import the MC parameter sets (Input parameters)
    filename = "InputParameters.csv"
    MCParametersets = pd.read_csv(refDataDir+filename)
    
    Inputs.addMarginals()
    Inputs.Marginals[0].Name = 'Permeability'
    Inputs.Marginals[0].InputValues = MCParametersets['Permeability']
    
    Inputs.addMarginals()
    Inputs.Marginals[1].Name = 'Porosity'
    Inputs.Marginals[1].InputValues = MCParametersets['Porosity']
    
    Inputs.addMarginals()
    Inputs.Marginals[2].Name = 'LeakWellPermeability'
    Inputs.Marginals[2].InputValues = MCParametersets['LeakWellPermeability']
    
    #Inputs.addMarginals()
    #Inputs.Marginals[0].Name = 'Permeability'
    #Inputs.Marginals[0].DistType = 'lognormal'
    #Inputs.Marginals[0].Parameters = [8.442534173750838e-13, 3.477725257585019e-12]
    
    #Inputs.addMarginals()
    #Inputs.Marginals[1].Name = 'Porosity'
    #Inputs.Marginals[1].DistType = 'lognormal'
    #Inputs.Marginals[1].Parameters = [0.15426630272993014, 0.09762653981615052]
    
    #Inputs.addMarginals()
    #Inputs.Marginals[2].Name = 'LeakWellPermeability'
    #Inputs.Marginals[2].DistType = 'lognormal'
    #Inputs.Marginals[2].Parameters = [1e-12, 4e-12]
    
    #=====================================================
    #======  POLYNOMIAL CHAOS EXPANSION METAMODELS  ======
    #=====================================================    
    MetaModelOpts = Metamodel(Inputs)

    # Specify the max degree to be compared by the adaptive algorithm:
    # The degree with the lowest Leave-One-Out cross-validation (LOO)
    # error (or the highest score=1-LOO)estimator is chosen as the final 
    # metamodel.
    MetaModelOpts.MinPceDegree = 1 # default = 1
    MetaModelOpts.MaxPceDegree = 4
    
    # q-quasi-norm 0<q<1 (default=1)
    MetaModelOpts.q = 1.0 #np.linspace(0.25,1, 4)
    
    # Select the sparse least-square minimization method for 
    # the PCE coefficients calculation:
    # 1)OLS: Ordinary Least Square  2)BRR: Bayesian Ridge Regression 
    # 3)LARS: Least angle regression  4)ARD: Bayesian ARD Regression
    # 5)SGDR: Stochastic gradient descent learning
    # 6)FastARD: Fast Bayesian ARD Regression
    # 7)VBL: Variational Bayesian Learning
    # 8)EBL: Emperical Bayesian Learning 
    MetaModelOpts.RegMethod = 'FastARD'
    
    # Print summary of the regression results
    # MetaModelOpts.DisplayFlag = True
    
    # ------ Experimental Design --------
    # Generate an experimental design of size NrExpDesign based on a latin 
    # hypercube sampling of the input model or user-defined values of X and/or Y:
    MetaModelOpts.addExpDesign()
    # MetaModelOpts.ExpDesign.MCSize = len(MCParametersets)
    
    # One-shot (normal) or Sequential Adaptive (normal) Design
    MetaModelOpts.ExpDesign.Method = 'normal'
    initNSamples = 10
    MetaModelOpts.ExpDesign.NrSamples = initNSamples

    # Sampling methods
    # 1) random 2) latin_hypercube 3) sobol 4) halton 5) hammersley 6) chebyshev(FT) 
    # 7) korobov 8) grid(FT) 9) nested_grid(FT) 10)user
    MetaModelOpts.ExpDesign.SamplingMethod = 'user'
    
    # Provide the experimental design object with a hdf5 file
    MetaModelOpts.ExpDesign.hdf5File = 'ExpDesign_CO2Benchmark.hdf5'
    
    # Sequential experimental design (needed only for sequential ExpDesign)
    MetaModelOpts.ExpDesign.NrofNewSample = 1
    MetaModelOpts.ExpDesign.MaxNSamples = 20
    MetaModelOpts.ExpDesign.ModifiedLOOThreshold = 1e-10
    
    # Defining the measurement error, if it's known a priori
    error = pd.DataFrame(np.ones((10)) * 1, columns=Model.Output.Names)
    
    DiscrepancyOpts = Discrepancy('')
    DiscrepancyOpts.Type = 'Gaussian'
    DiscrepancyOpts.Parameters = error **2
    MetaModelOpts.Discrepancy = DiscrepancyOpts
    
    # Plot the posterior snapshots for SeqDesign
    MetaModelOpts.ExpDesign.PostSnapshot = True
    MetaModelOpts.ExpDesign.stepSnapshot = 1
    MetaModelOpts.ExpDesign.MAP = (1.43e-13, 0.259, 1.09e-12)
    MetaModelOpts.ExpDesign.parNames = ["Permeability","Porosity","LeakWellPermeability"]
    
    
    # ------------------------------------------------
    # ------- Sequential Design configuration --------
    # ------------------------------------------------
    # 1) 'None' 2) 'equal' 3)'epsilon-decreasing' 4) 'adaptive'
    MetaModelOpts.ExpDesign.TradeOffScheme = 'None'
    #MetaModelOpts.ExpDesign.nReprications = 2
    # -------- Exploration ------
    #1)'Voronoi' 2)'random' 3)'dual annealing'
    MetaModelOpts.ExpDesign.ExploreMethod = 'random'
    
    # Use when 'dual annealing' chosen
    MetaModelOpts.ExpDesign.MaxFunItr = 200
    
    # Use when 'random'
    MetaModelOpts.ExpDesign.NCandidate = 100
    MetaModelOpts.ExpDesign.NrofCandGroups = 4
    
    # -------- Exploitation ------
    # 1)'BayesOptDesign' 2)'BayesActDesign' 3)'VarOptDesign' 4)'alphabetic' 
    # 5)'Space-filling'
    MetaModelOpts.ExpDesign.ExploitMethod = 'BayesActDesign'
    
    # BayesOptDesign -> when data is available
    # 1)DKL (Kullback-Leibler Divergence) 2)DPP (D-Posterior-percision)
    # 3)APP (A-Posterior-percision) 
    MetaModelOpts.ExpDesign.UtilityFunction = 'DKL' #['DKL', 'DPP']
    
    # VarBasedOptDesign -> when data is not available
    # Only with Vornoi >>> 1)Entropy 2)EIGF, 3)ALM, 4)LOOCV
    #MetaModelOpts.ExpDesign.UtilityFunction = 'Entropy'
    
    # alphabetic
    # 1)D-Opt (D-Optimality) 2)A-Opt (A-Optimality)
    # 3)K-Opt (K-Optimality)
    #MetaModelOpts.ExpDesign.UtilityFunction = 'D-Opt' #['D-Opt', 'A-Opt', 'K-Opt']
    
    # For colculation of validation error for SeqDesign 
    MetaModelOpts.validModelRuns = np.load(refDataDir+"origModelOutput.npy")
    MetaModelOpts.validSamples = np.load(refDataDir+"Prior.npy")
    MetaModelOpts.validLikelihoods = np.load(refDataDir+"validLikelihoods.npy")
    
    # >>>>>>>>>>>>>>>>>>>>>> Build Surrogate <<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<
    # Adaptive sparse arbitrary polynomial chaos expansion
    PCEModel = MetaModelOpts.create_metamodel(Model)
    
    
    #=====================================================
    #=========  POST PROCESSING OF METAMODELS  ===========
    #=====================================================
    PostPCE = PostProcessing(PCEModel)
    
    # Plot to check validation visually.
    PostPCE.validMetamodel(nValidSamples=3, x_axis="Time [day]")
    
    # Compute the moments and compare with the Monte-Carlo reference
    PostPCE.plotMoments()
    
    # Sensitivity analysis
    sobol_cell, total_sobol = PostPCE.sobolIndicesPCE()
    
    # Plot the evolution of the KLD,BME, and Modified LOOCV error
    #Information Entropy: -59.51615853207922
    if MetaModelOpts.ExpDesign.Method == 'sequential':
        PostPCE.seqDesignDiagnosticPlots(refBME_KLD=(-22.41279, 4.497334360207763))     
    #=====================================================
    #==============  Bayesian inference  =================
    #=====================================================
    BayesOpts = BayesInference(PCEModel)
    
    BayesOpts.emulator = True
    
    # Evaluation of surrogate model predictions
    BayesOpts.Samples = MCParametersets
    #BayesOpts.NrofSamples = 1000000
    
    # Bootstrap for BME calulations
    # BayesOpts.Bootstrap = True
    # BayesOpts.BootstrapItrNr = 1 
    
    # Select the inference method
    BayesOpts.SamplingMethod = 'MCMC'
    BayesOpts.MCMCnwalkers = 300
    import emcee
    moves=[(emcee.moves.DESnookerMove(), 0.1),
            (emcee.moves.DEMove(), 0.9*0.9),
            (emcee.moves.DEMove(gamma0=1.0), 0.9 * 0.1),
            ]
    BayesOpts.MCMCmoves = emcee.moves.StretchMove() #moves
    
    
    BayesOpts.PlotPostDist = True
    BayesOpts.PlotPostPred = True
    BayesOpts.Corner_title_fmt = '.2e'
    
    # ----- Define the discrepancy model -------
    error = pd.DataFrame(np.ones((10)) * 1, columns=Model.Output.Names)
    
    # (Option B) # 0.2% of the data
    DiscrepancyOpts = Discrepancy('')
    DiscrepancyOpts.Type = 'Gaussian'
    DiscrepancyOpts.Parameters =  error**2
    BayesOpts.Discrepancy = DiscrepancyOpts

    
    Bayes = BayesOpts.create_Inference()
    
    
    #=====================================================
    #==============  Save class objects  =================
    #=====================================================
    with open('CO2Benchmark_Results.pkl', 'wb') as output:
        pickle.dump(PCEModel, output, pickle.HIGHEST_PROTOCOL)
    
        pickle.dump(PostPCE, output, pickle.HIGHEST_PROTOCOL)
    
        pickle.dump(Bayes, output, pickle.HIGHEST_PROTOCOL)
    
#    del PCEModel
#    del PostPCE
#    del Bayes
    
    # Load the objects
#    with open('./CO2Benchmark_Results.pkl', 'rb') as input:
#        PCEModel = pickle.load(input)
#        PostPCE = pickle.load(input)
#        Bayes = pickle.load(input)
        
    
    
    #=====================================================
    #======= Compare Posterior distributions =============
    #=====================================================
    def posteriorPlot(Posterior, MAP, parNames, key, figsize=(10,10)):
        import corner
        newpath = (r'Outputs_PosteriorComparison') 
        if not os.path.exists(newpath): os.makedirs(newpath)
        
        nParams = len(parNames)
        figPosterior = corner.corner(Posterior, labels=parNames,title_fmt='.2e',
                           show_titles=True, title_kwargs={"fontsize": 12})
    
        # This is the true mean of the second mode that we used above:
        value1 = MAP
        
        # This is the empirical mean of the sample:
        value2 = np.mean(Posterior, axis=0)
    
        # Extract the axes
        axes = np.array(figPosterior.axes).reshape((nParams, nParams))
        
        # Loop over the diagonal
        for i in range(nParams):
            ax = axes[i, i]
            ax.axvline(value1[i], color="g")
            ax.axvline(value2[i], ls='--', color="r")
        
        # Loop over the histograms
        for yi in range(nParams):
            for xi in range(yi):
                ax = axes[yi, xi]
                ax.axvline(value1[xi], color="g")
                ax.axvline(value2[xi], ls='--', color="r")
                ax.axhline(value1[yi], color="g")
                ax.axhline(value2[yi], ls='--', color="r")
                ax.plot(value1[xi], value1[yi], "sg")
                ax.plot(value2[xi], value2[yi], "sr")
        
        figPosterior.set_size_inches((10,10)) 
        figPosterior.savefig('./'+newpath+'/'+key+'.svg',bbox_inches='tight')
    
    theta_true = MetaModelOpts.ExpDesign.MAP
    parNames = ["Permeability","Porosity","LeakWellPermeability"]
    # Extract the posterior with emulator
    surrogatePosterior = Bayes.Posterior_df
    
    # Plot surrogate posteriors
    posteriorPlot(surrogatePosterior, theta_true, parNames, 'surrogatePosterior')
    
    
    # Read the original posterior
    origPosterior = pd.read_csv(refDataDir+"origPosterior.csv")
    
    # Plot original posteriors
    posteriorPlot(origPosterior, theta_true, parNames, 'origPosterior')   
    
