#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Sat Oct 26 14:02:02 2019
http://jakevdp.github.io/blog/2014/06/14/frequentism-and-bayesianism-4-bayesian-in-python/
@author: farid
"""
import os, sys
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import scipy.stats as st
from scipy.linalg import cholesky as chol
import warnings
import corner
import shutil
import zipfile
try:
    import cPickle as pickle
except ModuleNotFoundError:
    import pickle
    
import matplotlib
matplotlib.use('agg')


# Create some convenience routines for plotting

def compute_sigma_level(trace1, trace2, nbins=20):
    """From a set of traces, bin by number of standard deviations"""
    L, xbins, ybins = np.histogram2d(trace1, trace2, nbins)
    L[L == 0] = 1E-16
    logL = np.log(L)

    shape = L.shape
    L = L.ravel()

    # obtain the indices to sort and unsort the flattened array
    i_sort = np.argsort(L)[::-1]
    i_unsort = np.argsort(i_sort)

    L_cumsum = L[i_sort].cumsum()
    L_cumsum /= L_cumsum[-1]
    
    xbins = 0.5 * (xbins[1:] + xbins[:-1])
    ybins = 0.5 * (ybins[1:] + ybins[:-1])

    return xbins, ybins, L_cumsum[i_unsort].reshape(shape)


def plot_MCMC_trace(emcee_trace, theta_truth, nameList, pointEstimate='Mean', **kwargs):
    """Plot traces and contours"""
    samples = emcee_trace.T#[:-1].T
    ndim = samples.shape[1]
    
    
    
    
    value1 = theta_truth
    if pointEstimate == 'Mean':
        value2 = np.mean(samples, axis=0)
    elif pointEstimate == 'MAP':
        value2 = st.mode(samples, axis=0)[0][0]
    
    try:
        figure = corner.corner(samples, labels=nameList)
    except:
        ranges = [(-5,5), (-5,5)]
        figure = corner.corner(samples, range=ranges, labels=nameList)
        
    # Extract the axes
    axes = np.array(figure.axes).reshape((ndim, ndim))
    
    # Loop over the diagonal
    for i in range(ndim):
        ax = axes[i, i]
        ax.axvline(value1[i], color="g")
        ax.axvline(value2[i], color="r")
    
    # Loop over the histograms
    for yi in range(ndim):
        for xi in range(yi):
            ax = axes[yi, xi]
            ax.axvline(value1[xi], color="g")
            ax.axvline(value2[xi], color="r")
            ax.axhline(value1[yi], color="g")
            ax.axhline(value2[yi], color="r")
            ax.plot(value1[xi], value1[yi], "sg")
            ax.plot(value2[xi], value2[yi], "sr")
    
    figure.set_size_inches((24,16)) 
            
    figure.savefig('./Poisterior_Distribution.svg',
                         bbox_inches='tight')
       
def plot_MCMC_model(ax, xdata, ydata, trace):
    """Plot the linear model and 2sigma contours"""
    ax.plot(xdata, ydata, 'ok')

    alpha, beta = trace[:2]
    xfit = np.linspace(-20, 120, 10)
    yfit = alpha[:, None] + beta[:, None] * xfit
    mu = yfit.mean(0)
    sig = 2 * yfit.std(0)

    ax.plot(xfit, mu, '-k')
    ax.fill_between(xfit, mu - sig, mu + sig, color='lightgray')

    ax.set_xlabel('x')
    ax.set_ylabel('y')


def plot_MCMC_results(xdata, ydata, trace, colors='k'):
    """Plot both the trace and the model together"""
    fig, ax = plt.subplots(1, 2, figsize=(10, 4))
    plot_MCMC_trace(ax[0], xdata, ydata, trace, True, colors=colors)
    plot_MCMC_model(ax[1], xdata, ydata, trace)
    
    
def my_gelman_rubin(x):
    """ Estimate the marginal posterior variance. Vectorised implementation. """
    m_chains, n_iters = x.shape

    # Calculate between-chain variance
    B_over_n = ((np.mean(x, axis=1) - np.mean(x))**2).sum() / (m_chains - 1)

    # Calculate within-chain variances
    W = ((x - x.mean(axis=1, keepdims=True))**2).sum() / (m_chains*(n_iters - 1))

    # (over) estimate of variance
    s2 = W * (n_iters - 1) / n_iters + B_over_n
    
    return s2

def my_ESS(x):
    """ 
    Compute the effective sample size of estimand of interest. 
    Vectorised implementation. 
    """
    m_chains, n_iters = x.shape

    variogram = lambda t: ((x[:, t:] - x[:, :(n_iters - t)])**2).sum() / (m_chains * (n_iters - t))

    post_var = my_gelman_rubin(x)

    t = 1
    rho = np.ones(n_iters)
    negative_autocorr = False

    # Iterate until the sum of consecutive estimates of autocorrelation is negative
    while not negative_autocorr and (t < n_iters):
        rho[t] = 1 - variogram(t) / (2 * post_var)

        if not t % 2:
            negative_autocorr = sum(rho[t-1:t+1]) < 0

        t += 1
    
    return int(m_chains*n_iters / (1 + 2*rho[1:t].sum()))

def iterative_scheme(N1, N2, q11, q12, q21, q22, r0, neff, tol, maxiter, criterion):
    """
    Iterative scheme as proposed in Meng and Wong (1996) to estimate the marginal likelihood    

    Parameters
    ----------
    N1 : TYPE
        DESCRIPTION.
    N2 : TYPE
        DESCRIPTION.
    q11 : TYPE
        DESCRIPTION.
    q12 : TYPE
        DESCRIPTION.
    q21 : TYPE
        DESCRIPTION.
    q22 : TYPE
        DESCRIPTION.
    r0 : TYPE
        DESCRIPTION.
    neff : TYPE
        DESCRIPTION.
    tol : TYPE
        DESCRIPTION.
    maxiter : TYPE
        DESCRIPTION.
    criterion : TYPE
        DESCRIPTION.

    Returns
    -------
    TYPE
        DESCRIPTION.

    """
    l1 = q11 - q12
    l2 = q21 - q22
    lstar = np.median(l1) # To increase numerical stability, 
                          # subtracting the median of l1 from l1 & l2 later
    s1 = neff/(neff + N2)
    s2 = N2/(neff + N2)
    r = r0
    r_vals = [r]
    logml = np.log(r) + lstar
    criterion_val = 1 + tol

    i = 0
    while (i <= maxiter) & (criterion_val > tol):
        rold = r
        logmlold = logml
        numi = np.exp(l2 - lstar)/(s1 * np.exp(l2 - lstar) + s2 * r)
        deni = 1/(s1 * np.exp(l1 - lstar) + s2 * r)
        if np.sum(~np.isfinite(numi))+np.sum(~np.isfinite(deni)) > 0:
            warnings.warn("""Infinite value in iterative scheme, returning NaN. 
            Try rerunning with more samples.""")
        r = (N1/N2) * np.sum(numi)/np.sum(deni)
        r_vals.append(r)
        logml = np.log(r) + lstar
        i += 1
        if criterion=='r':
            criterion_val = np.abs((r - rold)/r)
        elif criterion=='logml':
            criterion_val = np.abs((logml - logmlold)/logml)

    if i >= maxiter:
        return dict(logml = np.NaN, niter = i, r_vals = np.asarray(r_vals))
    else:
        return dict(logml = logml, niter = i)
        
def Marginal_llk_emcee(sampler, nburn=None, logp=None, maxiter=1000):
    """
    The Bridge Sampling Estimator of the Marginal Likelihood.

    Parameters
    ----------
    mtrace : MultiTrace, result of MCMC run
    model : PyMC Model
        Optional model. Default None, taken from context.
    logp : Model Log-probability function, read from the model by default
    maxiter : Maximum number of iterations

    Returns
    -------
    marg_llk : Estimated Marginal log-Likelihood.
    """
    r0, tol1, tol2 = 0.5, 1e-10, 1e-4
    
    if logp is None:
        logp = sampler.log_prob_fn

    
    # Split the samples into two parts  
    # Use the first 50% for fiting the proposal distribution and the second 50% 
    # in the iterative scheme.
    if nburn is None:
        mtrace = sampler.chain
    else:
        mtrace = sampler.chain[:, nburn:, :]
        
    nchain, len_trace, nrofVars = mtrace.shape
    
    N1_ = len_trace // 2
    N1 = N1_*nchain
    N2 = len_trace*nchain - N1
    
    samples_4_fit = np.zeros((nrofVars, N1))
    samples_4_iter = np.zeros((nrofVars, N2))
    effective_n = np.zeros((nrofVars))

    # matrix with already transformed samples
    for var in range(nrofVars):
        
        # for fitting the proposal
        x = mtrace[:,:N1_,var]
        
        samples_4_fit[var, :] = x.flatten()
        # for the iterative scheme
        x2 = mtrace[:,N1_:,var]
        samples_4_iter[var, :] = x2.flatten()

        # effective sample size of samples_4_iter, scalar 
        #(https://github.com/jwalton3141/jwalton3141.github.io/blob/master/assets/posts/ESS/rwmh.py)
        effective_n[var] = my_ESS(x2)

    # median effective sample size (scalar)
    neff = np.median(effective_n)
    
    # get mean & covariance matrix and generate samples from proposal
    m = np.mean(samples_4_fit, axis=1)
    V = np.cov(samples_4_fit)
    L = chol(V, lower=True)

    # Draw N2 samples from the proposal distribution
    gen_samples = m[:, None] + np.dot(L, st.norm.rvs(0, 1, 
                                         size=samples_4_iter.shape))

    # Evaluate proposal distribution for posterior & generated samples
    q12 = st.multivariate_normal.logpdf(samples_4_iter.T, m, V)
    q22 = st.multivariate_normal.logpdf(gen_samples.T, m, V)

    # Evaluate unnormalized posterior for posterior & generated samples
    import time
    start = time.time()
    q11 = np.asarray([logp(point) for point in samples_4_iter.T])
    q21 = np.asarray([logp(point) for point in gen_samples.T])
    
    end = time.time()
    multi_time = end - start
    print("Evaluate unnormalized posterior for posterior & generated samples"
          "took {0:.1f} seconds.".format(multi_time))
    
    # Run iterative scheme:
    tmp = iterative_scheme(N1, N2, q11, q12, q21, q22, r0, neff, tol1, maxiter, 'r')
    if ~np.isfinite(tmp['logml']):
        warnings.warn("""logml could not be estimated within maxiter, rerunning with 
                      adjusted starting value. Estimate might be more variable than usual.""")
        # use geometric mean as starting value
        r0_2 = np.sqrt(tmp['r_vals'][-2]*tmp['r_vals'][-1])
        tmp = iterative_scheme(q11, q12, q21, q22, r0_2, neff, tol2, maxiter, 'logml')

    return dict(logml = tmp['logml'], niter = tmp['niter'], method = "normal", 
                q11 = q11, q12 = q12, q21 = q21, q22 = q22)


def checkrange(theta, ranges): 
        c = False 
        #sigma = theta[-1]
        # traverse in the list1 
        for i, bounds in enumerate(ranges): 
            x = theta[i]
            # condition check 
            if x< bounds[0] or x> bounds[1]: #or sigma < 0: 
                c = True
        return c

def read_CO2(FileName):
    
    df = pd.read_csv(FileName)

    return df.to_numpy()[:,1]

def CO2Model(xx,Run_No):

    """
    CO2 benchmark model 
    
    """
    
    interface = "interface.xml"
    xx=np.array([xx])
    nParamSets, nParams = xx.shape
    
    # dirname
    nRun = Run_No
    dirname = 'CO2Benchmark_'+str(nRun)
    
    # Create a directory
    if not os.path.exists(dirname): os.makedirs(dirname,exist_ok = True)
    
    # Enter the dir    
    os.chdir(dirname)
    
    # Copy the interface 
    shutil.copy2('../'+interface, '.')
    
    # Save the parameterset xx as .npy
    np.save('params.npy',xx)
    
    # Replace parameters
    text_to_search_list = ['<X%s>'%(i+1) for i in range(nParams)]
    with open(interface, 'r') as file :
        filedata = file.read()
        
        # Replace the target string
        for text_to_search, CollocationPoint in zip(text_to_search_list, xx[0]):
            filedata = filedata.replace(text_to_search, '{:0.4e}'.format(CollocationPoint))
        
        # Write the file out again
        with open(interface, 'w') as file:
          file.write(filedata)
    
    # Run the model
    NewCommand = "model_CO2.py -f %s"%interface
    Process = os.system('./../%s' %NewCommand)
    
    # Read the output
    Output = read_CO2('PressureOutput.csv')
    
    os.chdir('..')
    
    return Output


        
if __name__ == '__main__':
    cnt = None
    def init(args):
        ''' store the counter for later use '''
        global cnt
        cnt = args
    
    
    # --------------------------------------------
    # ----------- Define parameters --------------
    # --------------------------------------------
    X1 = (8.442534173750838e-13, 3.477725257585019e-12) 
    X2 = (0.15426630272993014, 0.09762653981615052)
    X3 = (1e-12, 4e-12) 
    
    params_moments=[]
    for params in [X1, X2, X3]:
        Mu = np.log(params[0]**2 / np.sqrt(params[0]**2 + params[1]**2))
        Sigma  = np.sqrt(np.log(1 + params[1]**2 / params[0]**2))
        params_moments.append((Mu,Sigma))
    
    params_range = [(1.718968e-16,2.848961e-11),(0.001107,0.415985),
                    (4.247943e-13, 2.984176e-12)]

    # --------------------------------------------
    # ----------- Define observation -------------
    # --------------------------------------------
    np.random.seed(42)

    theta_true = (1.43e-13, 0.259, 1.09e-12)
    ydata = np.array([326.2244181083123, 313.78768336474246, 315.0499230817254,
                      313.2715183529026, 314.20008286525774, 312.1947579212001,
                      313.12579464097877, 309.0260453794458, 312.30845007153476,
                      312.4797560102911])
    
    nObs = ydata.shape[0]
    
    # -------------------------------------------------------------------------
    # ------------------- MCMC Affine Invarient with emcee package ------------
    # -------------------------------------------------------------------------
    # Define our posterior using Python functions
    # Note that emcee requires log-posterior
    import chaospy
    import emcee
    
    def log_prior(theta):
        global params_moments, params_range
        
        #X1, X2, sigma = theta
        X1, X2, X3 = theta
        if checkrange(theta, params_range):
            return -np.inf  # log(0)
        else:
            # lognormal prior on X1, X2, X3
            ln_X1 = np.log(chaospy.LogNormal(mu=params_moments[0][0], sigma=params_moments[0][1]).pdf(X1))
            ln_X2 = np.log(chaospy.LogNormal(mu=params_moments[1][0], sigma=params_moments[1][1]).pdf(X2))
            ln_X3 = np.log(chaospy.LogNormal(mu=params_moments[2][0], sigma=params_moments[2][1]).pdf(X3))
            
            
            #ln_sigma = st.halfcauchy.logpdf(sigma, loc=0, scale = 3)
            
            return ln_X1 + ln_X2 + ln_X3 #+ ln_sigma
    
    def log_likelihood(theta, y, Run_No):
        
        
        # Build covariance matrix
        sigma = 1 #theta[-1]
        cov_matrix = np.diag(np.repeat(sigma**2, y.shape[0]))
        
        y_model = CO2Model(theta,Run_No+1) #[:-1])
        
        # handle non-converged simulations
        if np.isnan(np.min(y_model)):
            return (Run_No, -np.inf)
        else:
            return (Run_No, st.multivariate_normal.logpdf(y_model, mean=y, cov=cov_matrix))
        
    def log_posterior(theta, y):
        if log_prior(theta) == -np.inf : 
            return -np.inf
        else:
            return log_prior(theta) + log_likelihood(theta, y)
    
    
    # Here we'll set up the computation. emcee combines multiple "walkers",
    # each of which is its own MCMC chain. The number of trace results will
    # be nwalkers * nsteps
    
    ndim = len(params_range)  # number of parameters in the model
    nwalkers = 10  # number of MCMC walkers
    nburn = 100 # "burn-in" period to let chains stabilize
    nsteps = 1000  # number of MCMC steps to take
    ntemps = 20
    
    # set theta near the maximum likelihood, with 
    np.random.seed(0)
    starting_guesses = np.zeros((nwalkers, ndim))
    
    for idxDim in range(ndim):
        lower, upper = params_range[idxDim]
        starting_guesses[:,idxDim] = st.uniform(loc=lower, scale=upper-lower).rvs(size=nwalkers)

    #sigma
    #starting_guesses[:,ndim-1] = st.halfcauchy(loc=0, scale = 1).rvs(size=nwalkers)
    
    
    # -------------------------------------------------------------------------
    # ------------------------ MCMC with emcee package ------------------------
    # -------------------------------------------------------------------------
    # from multiprocessing import Pool, Value,Lock
    # counter = Value('i', 0)
    # counter_lock = Lock()
    
    # with Pool() as pool:
    #     sampler = emcee.EnsembleSampler(nwalkers, ndim, log_posterior, args=[ydata], pool=pool)
    #     sampler.run_mcmc(starting_guesses, nsteps, progress=True)
        
    # with open('CO2Benchmark_emcee.pkl', 'wb') as output:
    #     pickle.dump(sampler, output, pickle.HIGHEST_PROTOCOL)
        
    # Load the objects
    # with open('Results_05-05-2020_10-16/CO2Benchmark_emcee.pkl', 'rb') as input:
    #     sampler = pickle.load(input)

    # # sampler.chain is of shape (nwalkers, nsteps, ndim)
    # # we'll throw-out the burn-in points and reshape:
    # emcee_trace = sampler.chain[:, nburn:, :].reshape(-1, ndim).T
    
    # # Plot with corner and the truth
    # nameList = ['Permeability', 'Porosity','LeakWellPermeability']
    
    # plot_MCMC_trace(emcee_trace, theta_true, nameList, pointEstimate='Mean')
    
    # logml_dict = Marginal_llk_emcee(sampler, nburn, logp=None, maxiter=5000)
    # print('\nThe Bridge Sampling Estimation is %.5f.'%(logml_dict['logml']))
       
    
    # # Posterior-based expectation of posterior probablity
    # postLikelihoods = sampler.get_log_prob(flat=True)[nburn*nwalkers:]
    # postExpPostLikelihoods_emcee = np.mean(postLikelihoods)
    
    # # Posterior-based expectation of prior densities
    # priorLikelihoods = [log_prior(sample) for sample in emcee_trace.T]
    # postExpPrior_emcee = np.mean(priorLikelihoods)
    
    # # Posterior-based expectation of likelihoods
    # postExpLikelihoods_emcee = postExpPostLikelihoods_emcee - postExpPrior_emcee
    
    # # Calculate Kullback-Leibler Divergence
    # KLD_emcee = postExpLikelihoods_emcee - logml_dict['logml']
    # print("Kullback-Leibler divergence: %.5f"%KLD_emcee)
    
    # # Information Entropy based on Entropy paper Eq. 38
    # infEntropy_emcee = logml_dict['logml'] - postExpPrior_emcee - postExpLikelihoods_emcee
    # print("Information Entropy: %.5f" %infEntropy_emcee)
    
    # -------------------------------------------------------------------------
    # ------------------ Naive Monte-Carlo with rejection sampling ------------
    # -------------------------------------------------------------------------
    mcSize = 10000
    Likelihood = np.zeros((mcSize))
    prior = np.zeros((mcSize,ndim))
    
    for idxDim in range(ndim):
        Params = params_moments[idxDim]
        ranges = params_range[idxDim]
        prior[:,idxDim] = chaospy.LogNormal(mu=Params[0], sigma=Params[1]).sample(mcSize,rule="random")
        
        # Check the ranges if not in range shift them.
        while any(prior[:,idxDim] < ranges[0]) or any(prior[:,idxDim] > ranges[1]):
            prior[:,idxDim][prior[:,idxDim] < ranges[0]] +=ranges[0]
            prior[:,idxDim][prior[:,idxDim] > ranges[1]] -= ranges[1]
            
    
    from multiprocessing import Pool
    
    pool = Pool()
        
    # With Pool.starmap_async()
    results = pool.starmap_async(log_likelihood, [(theta_i, ydata, i) for i, theta_i in enumerate(prior)]).get()
        
    # Close the pool
    pool.close()
    
    # Collect the results from pull
    Likelihood = np.exp([results[NofE][1] for NofE in range(mcSize)])
    print("Likelihood:\n",Likelihood)
    
    logml = np.log(np.nanmean(Likelihood))
    print('\nThe Naive MC-Estimation of BME is %.5f.'%(logml))
    
    normLikelihood = Likelihood / np.nanmax(Likelihood)
    # Random numbers between 0 and 1
    unif = np.random.rand(1, mcSize)[0]
    
    # Reject the poorly performed prior
    accepted = normLikelihood >= unif
    acceptedSamples = prior[accepted]
    
    # Plot with corner and the truth
    nameList = ['Permeability', 'Porosity','LeakWellPermeability']
    NaiveMC_trace = acceptedSamples.T
    plot_MCMC_trace(NaiveMC_trace, theta_true, nameList, pointEstimate='Mean')
    
    
    # -------------------------------------------------------------------------
    # ------------- Kullback-Leibler Divergence & Information Entropy ---------
    # -------------------------------------------------------------------------
    # Prior-based estimation of BME
    logBME = logml
    
    # Posterior-based expectation of likelihoods
    postExpLikelihoods = np.mean(np.log(Likelihood[accepted]))
    
    # Posterior-based expectation of prior densities
    postExpPrior = np.mean([log_prior(sample) for sample in acceptedSamples])
    
    
    # Calculate Kullback-Leibler Divergence
    KLD = postExpLikelihoods - logBME
    print("Kullback-Leibler divergence:", KLD)
    
    # Information Entropy based on Entropy paper Eq. 38
    infEntropy = logBME - postExpPrior - postExpLikelihoods
    print("Information Entropy:", infEntropy)
    
    # -------------------------------------------------------------------------
    # ------------------------- Save necessary outputs ------------------------
    # -------------------------------------------------------------------------
    nonzeroLikelihood = Likelihood !=0
    
    # Posterior as csv
    np.savetxt("InputParameters.csv", prior, fmt='%.8e', delimiter=',',
               header='Permeability,Porosity,LeakWellPermeability', comments='')
    
    # Likelihoods as npy
    np.save('validLikelihoods.npy', np.log(Likelihood[nonzeroLikelihood]))
    
    # Posterior as csv
    np.savetxt("origPosterior.csv", NaiveMC_trace, fmt='%.8e', delimiter=',',
               header='Permeability,Porosity,LeakWellPermeability', comments='')
    
    # Prior and Model outputs as npy
    Prior = np.empty((0,ndim))
    TotalOutputs = np.empty((0,nObs))
    folderPaths = [path for path in os.listdir('.') if 'CO2Benchmark_' in path]
    sortedfolderPaths = sorted(folderPaths, key=lambda x: int("".join([i for i in x if i.isdigit()])))
    
    for folder in sortedfolderPaths:
        output = pd.read_csv(folder+'/PressureOutput.csv').to_numpy()[:,1]
        if not np.isnan(np.min(output)):
            Prior = np.vstack((Prior,np.load(folder+'/params.npy')))
            TotalOutputs = np.vstack((TotalOutputs,output))
    
    np.save('Prior.npy', Prior)
    np.save('origModelOutput.npy', TotalOutputs)
    
    # Mean and Std of outputs as refs_MC csv
    timeVector = pd.read_csv(folder+'/PressureOutput.csv').to_numpy()[:,0]
    mean = np.mean(TotalOutputs, axis=0)
    std = np.std(TotalOutputs, axis=0)
    
    np.savetxt('MCrefs_MeanStd.csv', np.vstack((timeVector,mean,std)).T, fmt='%.8f',
               delimiter=',', header='Time [day],mean,std', comments='')
    
    # Zip folders
    zf = zipfile.ZipFile("ModelRuns.zip", "w")
    for dirname, subdirs, files in os.walk("."):
        if 'CO2Benchmark_' in dirname:
            zf.write(dirname)
            for filename in files:
                zf.write(os.path.join(dirname, filename))
            shutil.rmtree(dirname)
    zf.close()
    
